"""Top-level module for Wtforg Library."""

import subprocess
import sys
from typing import Tuple, Union

from wtforglib.options import Options


class FakedProcessResult(object):
    """Faked process result."""

    stdout: str
    stderr: str
    returncode: int

    def __init__(self, stdout: str = "", stderr: str = "", returncode: int = 0) -> None:
        """Creates a fake process result.

        Parameters
        ----------
        stdout : str
            Fake stdout
        stderr : str
            Fake stderr
        returncode : int
            Fake returcode
        """
        self.stdout = stdout
        self.stderr = stderr
        self.returncode = returncode


if sys.version_info >= (3, 9):
    CommanderResult = Union[subprocess.CompletedProcess[str], FakedProcessResult]
else:
    CommanderResult = Union[subprocess.CompletedProcess, FakedProcessResult]


class Commander(Options):
    """Base class for subprocess management."""

    def run_command(
        self,
        args: Tuple[str, ...],
        **kwargs: bool,
    ) -> CommanderResult:
        """Runs commands specified by args."""
        always = kwargs.get("always", False)
        check = kwargs.get("check", True)
        cmd_str = "{0}".format(" ".join(args))
        if not always and self.isnoop():
            print("noex: {0}".format(cmd_str))
            return FakedProcessResult()
        self.verbose("ex: {0}".format(cmd_str), 2)
        return subprocess.run(
            args,
            check=check,
            shell=False,  # noqa: S603
            capture_output=True,
            encoding="utf-8",
        )



"""Top-level module for wtforglib Library."""
import platform
import subprocess  # noqa: S404
from typing import ClassVar

from wtforglib.errors import ShellError

WINDOZE = "Windows"

if platform.system() == WINDOZE:
    from socket import getfqdn  # noqa: WPS433

WINDOZE_NOT_IMPLEMENTED = "{0} is not implemented on Windows"


class WtfSingleton(object):
    """Singleton class."""

    dn: ClassVar[str] = ""
    hn: ClassVar[str] = ""

    def __new__(cls):  # pragma no cover
        """Creates new singleton if one does not exist."""
        if not hasattr(cls, "instance"):  # noqa: WPS421
            cls.instance = super(WtfSingleton, cls).__new__(cls)  # noqa: WPS608
        return cls.instance


def domainname(test: bool = False) -> str:  # noqa: WPS605, WPS231
    """Retuns hosts domainname.

    Parameters
    ----------
    test : bool Optional
        Default False, when true return example.com

    Returns
    -------
    str
        domainname

    Raises
    ------
    ShellError
        if subprocess returncode is not 0
    """
    if not WtfSingleton.dn:
        if test:
            WtfSingleton.dn = "example.com"
        elif platform.system() == WINDOZE:
            parts = getfqdn().split(".", 1)
            if len(parts) == 2:
                WtfSingleton.dn = parts[1]
            else:
                WtfSingleton.dn = "unknown"
        else:
            sp_result = subprocess.run(
                ["hostname", "-d"],
                capture_output=True,
                encoding="utf8",
            )
            if sp_result.returncode != 0:  # pragma no cover
                raise ShellError("Failed to determine host's domainname")
            WtfSingleton.dn = sp_result.stdout.rstrip()
    return WtfSingleton.dn


def hostname(test: bool = False) -> str:  # noqa: WPS605, WPS231
    """Retuns hosts hostname.

    Parameters
    ----------
    test : bool Optional
        Default False, when true return nombre

    Returns
    -------
    str
        hostname

    Raises
    ------
    ShellError
        if subprocess returncode is not 0
    """
    if not WtfSingleton.hn:
        if test:
            WtfSingleton.hn = "nombre"
        elif platform.system() == WINDOZE:
            parts = getfqdn().split(".", 1)
            if parts:
                WtfSingleton.hn = parts[0]
            else:
                WtfSingleton.hn = "unknown"
        else:
            sp_result = subprocess.run(
                ["hostname"],
                capture_output=True,
                encoding="utf8",
            )
            if sp_result.returncode != 0:  # pragma no cover
                raise ShellError("Failed to determine hostname")
            WtfSingleton.hn = sp_result.stdout.rstrip()
    return WtfSingleton.hn


def strtobool(rts: str) -> bool:
    """Covert string rts to boolean."""
    return rts.lower() in {"true", "1", "t", "y", "yes"}


def windoze_not_implemented(foo_name: str) -> None:
    """Raised and exception if platform is Windows.

    Parameters
    ----------
    foo_name : str
        Name of function not supported

    Raises
    ------
    NotImplementedError
        When platform is Windows
    """
    if platform.system() == WINDOZE:
        raise NotImplementedError(
            WINDOZE_NOT_IMPLEMENTED.format(foo_name),
        )  # pragma: no cover



import os
import re
import shutil


def clear_slot(root: str, idx: int, max_versions: int, debug: bool = False) -> str:
    """Clear file out of slot.

    Parameters
    ----------
    root : str
        un-numbered basename
    idx : int
        version number
    max_versions : int
        maximum number of versions
    debug : bool, optional
        debug flag, by default False

    Returns
    -------
    str
        numbered slot name
    """
    slot = "{0}.{1}".format(root, idx)
    if os.path.isfile(slot):
        if idx >= max_versions:
            if debug:  # pragma no cover
                print("unlinking slot {0}".format(slot))
            os.unlink(slot)
        else:
            if debug:  # pragma no cover
                print("clearing slot {0}".format(slot))
            nslot = clear_slot(root, idx + 1, max_versions, debug)
            os.rename(slot, nslot)
    return slot


def check_root_filename(file_spec: str) -> str:
    """Determine root filename so the extension doesn't get longer.

    Parameters
    ----------
    file_spec : str
        Path name of the file to check

    Returns
    -------
    str
        Path name of the file to backup

    Raises
    ------
    ValueError
        If file_spec ends with one or more digit extension
    """
    nn, ee = os.path.splitext(file_spec)

    if re.match(r".\d+$", ee):
        raise ValueError(
            "Cannot create numbered backups for a file with a numbered ext",
        )
    return file_spec


def version_file(
    file_spec: str,
    vtype: str = "rename",
    max_versions: int = 5,
    debug: bool = False,
) -> int:
    """Save max versions of file.

    Parameters
    ----------
    file_spec : str
        Path to the file to be versioned.
    vtype : str, optional
        Either rename or copy when versioning, by default "rename"
    max_versions : int, optional
        maximum number of versions, by default 5
    debug : bool, optional
        debug flag, by default False

    Returns
    -------
    int
        exit code
    """
    if not os.path.isfile(file_spec):  # pragma no cover
        return 1
    # or, do other error checking:
    if vtype not in {"copy", "rename"}:  # pragma no cover
        vtype = "rename"

    root = check_root_filename(file_spec)

    # Find next available file version
    new_file = clear_slot(root, 1, max_versions, debug)
    # the code below is reported as not covered, but I
    # have ran severl tests to verify, I suspect, I need
    # use a fake file system for testing but not now
    if not os.path.isfile(new_file):  # pragma no cover
        if vtype == "copy":
            shutil.copy(file_spec, new_file)
        else:
            os.rename(file_spec, new_file)

    return 0


# vim:ft=py noqa: E800



"""Top-level module for wtforglib Library."""
import json
from pathlib import Path
from tempfile import TemporaryFile
from typing import Optional, Tuple

import yaml

from wtforglib.errors import raise_filenotfound
from wtforglib.kinds import Fspec, StrAnyDict


def load_yaml_file(
    filenm: Fspec,
    missing_ok: bool = True,
) -> StrAnyDict:
    """Loads a yaml file.

    Parameters
    ----------
    filenm : Fspec
        The yaml file to load
    missing_ok : bool optional default True
        When False and file does not exist raises FileNotFound exeception

    Returns
    -------
    StrAnyDict
        Representing contents of filenm
    """
    ypath = Path(filenm)
    yaml_result = {}
    if ypath.exists() and ypath.is_file():
        with open(ypath, "r") as yfile:
            yaml_result = yaml.safe_load(yfile)
    elif not missing_ok:  # pragma no cover
        raise_filenotfound(ypath)
    return yaml_result


def write_yaml_file(
    filenm: Fspec,
    src_data: StrAnyDict,
    encoding: str = "utf-8",
) -> bool:
    """Writes src_data to a file in a yaml format.

    Parameters
    ----------
    filenm : Fspec
        The yaml filename to write to
    src_data : StrAnyDict
        The data to write to a file
    encoding : str Optional
        encoding default 'utf-8'

    Returns
    -------
        bool
            True if file exists else False
    """
    ypath = Path(filenm)
    with open(ypath, "w") as outf:
        yaml.dump(src_data, outf, encoding=encoding)
    return ypath.exists()


def load_json_file(
    filenm: Fspec,
    missing_ok: bool = True,
) -> StrAnyDict:
    """Loads a json file.

    Parameters
    ----------
    filenm : Fspec
        The yaml file to load
    missing_ok : bool optional default True
        When False and file does not exist raises FileNotFound exeception

    Returns
    -------
    StrAnyDict
        Representing contents of filenm
    """
    jpath = Path(filenm)
    rtn: StrAnyDict = {}
    if jpath.exists() and jpath.is_file():
        with open(jpath, "r") as jfile:
            rtn = json.load(jfile)
    elif not missing_ok:  # pragma no cover
        raise_filenotfound(jpath)
    return rtn


def write_json_file(
    filenm: Fspec,
    src_data: StrAnyDict,
    indent: int = 2,
) -> bool:
    """Writes src_data to a file in a json format.

    Parameters
    ----------
    filenm : Fspec
        The json filename to write to
    src_data : StrAnyDict
        The data to write to a file
    indent : int Optional
        The number of spaces to indent

    Returns
    -------
        bool
            True if file exists else False
    """
    jpath = Path(filenm)
    with open(jpath, "w") as outf:
        outf.write(json.dumps(src_data, indent=2))
    return jpath.exists()


def ensure_directory(target: Fspec, perm: int = 0o755) -> bool:
    """Creates a directory if it doesn't exist.

    Parameters
    ----------
    target : Fspec
        The path to the target directory
    perm : int Optional
        The mode to use if creationg needed

    Returns
    -------
    bool
        True if directory exists

    Raises
    ------
    NotADirectoryError
        if target exists, but not a directory
    """
    dp = Path(target)
    if dp.is_dir():
        return True
    if dp.exists():
        raise NotADirectoryError(
            'Pathname "{0}" is not a directory.'.format(str(target)),
        )
    dp.mkdir(mode=perm, parents=True)
    return True


def _verify_directory_write(dpath: Fspec) -> Optional[Exception]:
    """Verify that the given directory is writable.

    Parameters
    ----------
    dpath : Fspec
        Pathlike object specifying the directory

    Returns
    -------
    Optional[Exception]
        The exception when caught
    """
    error: Optional[Exception] = None
    try:
        tfile = TemporaryFile(dir=dpath)
        tfile.close()
    except Exception as ex:
        error = ex
    return error


def verify_directory(dspec: Fspec, ex: bool = False) -> Tuple[bool, str]:
    """Verify that a directory exits and is writable.

    Parameters
    ----------
    dspec : Fspec
        Pathlike object specifying the directory
    ex : bool, optional
        When True exceptions are raised, by default False

    Returns
    -------
    Tuple[bool, str]
        When ex is False the status is returned

    Raises
    ------
    error
        NotADirectoryError, FileNotFoundError or other exception is raised on failure
        when ex is True
    """
    dpath = Path(dspec)
    dstr = str(dpath)
    error: Optional[Exception] = None
    if dpath.exists():
        if dpath.is_dir():
            error = _verify_directory_write(dpath)
        else:
            error = NotADirectoryError("'{0}' is not a directory".format(dstr))
    else:
        error = FileNotFoundError("Directory not found: {0}".format(dpath))
    if error:
        if ex:
            raise error
    return (error is None, str(error))



"""Top-level module for wtforglib."""

import sys
from pathlib import Path

from wtforglib.kinds import Fspec


def unlink_path(fspec: Fspec, missing_ok: bool = False) -> None:
    """Unlink file with version support.

    Parameters
    ----------
    fspec : Fspec
        Specification of file to unlink
    missing_ok : bool
        Flag to specify if file must exist

    Raises
    ------
    FileNotFoundError
        If missing_ok is False and file does not exist.
    """
    path = Path(fspec)
    if sys.version_info >= (3, 8):
        path.unlink(missing_ok=missing_ok)
    else:
        try:
            path.unlink()
        except FileNotFoundError:
            if not missing_ok:
                raise



"""Top-level module for Wtforg Library."""

import errno
import os
from pathlib import Path
from typing import NoReturn

from wtforglib.kinds import Fspec


class ShellError(Exception):
    """Raised when subprocess resultcode not 0."""


def raise_filenotfound(filenm: Fspec) -> NoReturn:
    """Raises a FileNotFoundError execption for the given filenm.

    Parameters
    ----------
    filenm : Union[str,PathLike[str]]
        The yaml file to load

    Raises
    ------
    FileNotFoundError
        FileNotFoundError
    """
    raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), str(filenm))


def raise_filenotfound_if(filenm: Fspec) -> None:
    """Call raise_filenotfound for the given filenm if it doesn't exist.

    Parameters
    ----------
    filenm : Fspec
        Name of the file to test
    """
    fp = Path(filenm)
    if not fp.exists():
        raise_filenotfound(filenm)
    if not fp.is_file():
        raise_filenotfound(filenm)



"""Top-level module for Wtforg Library."""

from typing import Dict, Optional, Union

from wtforglib.scribe import LOG_LEVELS, Scribe

OptionsDict = Dict[str, Union[str, int, bool]]

OPTIONS_DEBUG = "debug"
OPTIONS_TEST = "test"
OPTIONS_VERBOSE = "verbose"


def log_level_name(level: int, default: str = "info") -> str:
    """Return logger level name.

    Parameters
    ----------
    level : int
        Level number
    default : str, optional
        Default if not matched, by default "info"

    Returns
    -------
    str
        Level name
    """
    for key, valor in LOG_LEVELS.items():
        if level == valor:
            return key
    return default


def basic_options(
    debug: Union[bool, int],
    test: bool,
    verbose: Union[bool, int],
) -> OptionsDict:
    """Return a dictionary of basic options.

    Parameters
    ----------
    debug : Union[bool, int]
        Sets debug: int in the returned dictionary
    test : bool
        Sets test: bool in the returned dictionary
    verbose : Union[bool, int]
        Sets verbose: int in the returned dictionary

    Returns
    -------
    OptionsDict
        Options dictionary
    """
    od: OptionsDict = {}
    od["debug"] = int(debug) if isinstance(debug, bool) else debug
    od["test"] = test
    od["verbose"] = int(verbose) if isinstance(verbose, bool) else verbose
    return od


class Options(object):  # noqa: WPS214
    """A class to handle Options."""

    options: OptionsDict
    scribe: Optional[Scribe]

    def __init__(self, opts: Optional[OptionsDict] = None) -> None:
        """Options constructor.

        Class to manage options for the module.

        Parameters
        ----------
        opts : OptionsDict Optional
            Options to merge with defaults
        """
        self.options = {OPTIONS_DEBUG: 0, OPTIONS_VERBOSE: 0, OPTIONS_TEST: False}
        self.errors = 0
        self.scribe = None
        if opts is not None:  # pragma no cover ???
            self.options.update(opts)
        if int(self.options[OPTIONS_DEBUG]) > 1:
            print(  # noqa: WPS421
                "created instance of class {0}".format(self.__class__.__name__),
            )

    def isdebug(self) -> bool:
        """Returns True if debug > 0."""
        return int(self.options[OPTIONS_DEBUG]) > 0

    def isverbose(self) -> bool:
        """Returns True if verbose > 0."""
        return int(self.options[OPTIONS_VERBOSE]) > 0

    def isnoop(self) -> bool:
        """Returns True if noop."""
        return bool(self.options.get("noop", False))

    def isforce(self) -> bool:
        """Returns True if noop."""
        return bool(self.options.get("force", False))

    def istest(self) -> bool:
        """Returns value of test flag."""
        return bool(self.options[OPTIONS_TEST])

    def debug(self, message: str, level: int = 1) -> None:
        """Utility debug method."""
        dl: int = int(self.options[OPTIONS_DEBUG])
        if self.options[OPTIONS_TEST] or dl >= level:  # pragma no branch
            self._dtrace(message)

    def verbose(self, message: str, level: int = 1) -> None:
        """Utility verbose method."""
        vl: int = int(self.options[OPTIONS_VERBOSE])
        if self.options[OPTIONS_TEST] or vl >= level:  # pragma no cover ???
            self._itrace(message)

    def info(self, message: str) -> None:  # noqa: WPS110
        """Utility trace info method."""
        if self.scribe is None:
            print("INFO: {0}".format(message))
        else:
            self.scribe.logger.info(message)

    def warning(self, message: str) -> None:
        """Utility trace warning method."""
        if self.scribe is None:
            print("WARNING: {0}".format(message))
        else:
            self.scribe.logger.warning(message)

    def warn(self, message: str) -> None:
        """Utility trace warn method."""
        self.warning(message)

    def error(self, message: str) -> None:
        """Utility trace error method."""
        if self.scribe is None:
            print("ERROR: {0}".format(message))
        else:
            self.scribe.logger.error(message)
        self.errors += 1

    def _dtrace(self, message: str) -> None:
        """Utility debug trace method."""
        if self.scribe is None:
            print("DEBUG: {0}".format(message))
        else:
            self.scribe.logger.debug(message)

    def _itrace(self, message: str) -> None:
        """Utility info trace method."""
        if self.scribe is None:
            print("INFO: {0}".format(message))
        else:
            self.scribe.logger.info(message)



"""Top-level module for wtforglib Library."""

from ipaddress import IPv6Address, IPv6Network, ip_address, ip_network

MAX_PREFIX_LENGTH6 = 128


def _get_ipv6_prefix(ipv6_addr: str, prefix_len: int) -> str:  # noqa: WPS210
    """Retruns prefix of ipv6 address.

    Parameters
    ----------
    ipv6_addr : str
        The ipv6 address
    prefix_len : int
        The length of the prefix

    Returns
    -------
    str
        The ipv6 network prefix

    Raises
    ------
    ValueError
        When the prefix length is not valid
    """
    valid_lengths = (16, 32, 48, 64, 80, 96, 112, 128)
    if prefix_len not in valid_lengths:
        raise ValueError(
            "Invalid prefix length: {0}. Expected one of {1}".format(
                prefix_len,
                ",".join(str(xx) for xx in valid_lengths),
            ),
        )
    prefix = ""
    cur_prefix_len = 0
    prefix_increment = 16
    ipv6_parts = ipv6_addr.split(":")
    for part in ipv6_parts:
        if prefix:
            prefix = "{0}:{1}".format(prefix, part)
        else:
            prefix = part
        cur_prefix_len += prefix_increment
        if int(cur_prefix_len) >= int(prefix_len):
            return prefix
    return prefix  # pragma no cover


def is_ipv6_address(ipaddr: str) -> bool:
    """Returns True if ipaddr is a valid IPv6 address.

    Parameters
    ----------
    ipaddr : str
        A string representation of a IPv6 address

    Returns
    -------
    bool
        True if ipaddr is a valid IPv6 address
    """
    try:
        osix = ip_address(ipaddr)
        return isinstance(osix, IPv6Address)
    except ValueError:
        return False


def is_ipv6_network(ipaddr: str) -> bool:
    """Returns True if ipaddr is a valid IPv6 network.

    Parameters
    ----------
    ipaddr : str
        A string representation of a IPv6 network

    Returns
    -------
    bool
        True if ipaddr is a valid IPv6 network
    """
    try:
        osix = ip_network(ipaddr)
        return isinstance(osix, IPv6Network)
    except ValueError:
        return False


def is_ipv6(ipvalue: str) -> bool:
    """Returns True if ipvalue is a valid IPv6 address or IPv6 network.

    Parameters
    ----------
    ipvalue : str
        A string representation of a IPv6 address or IPv6 network

    Returns
    -------
    bool
        True if ipvalue is a valid IPv6 address or IPv6 network
    """
    if not is_ipv6_address(ipvalue):
        return is_ipv6_network(ipvalue)
    return True


def ipv6_to_netprefix(ipaddr: str, prefix_len: int) -> str:
    """Returns ipv6 network prefix.

    Parameters
    ----------
    ipaddr : str
        The ip address
    prefix_len : int
        The network prefix length

    Returns
    -------
    str
        ipv6 network prefix Example: '1111:2222:3333:4444::/64'

    Raises
    ------
    ValueError
        When ipaddr is not a valid ipv6 address
    """
    ipobj = ip_address(ipaddr)
    if isinstance(ipobj, IPv6Address):
        if prefix_len == MAX_PREFIX_LENGTH6:
            return "{0}/{1}".format(str(ipobj), prefix_len)
        return "{0}::/{1}".format(_get_ipv6_prefix(ipaddr, prefix_len), prefix_len)
    raise ValueError("Invalid ipv6 address: {0}".format(ipaddr))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['wtforglib']

package_data = \
{'': ['*']}

install_requires = \
['jinja2>=3.1.2,<4.0.0',
 'pyyaml>=6.0,<7.0',
 'types-pyyaml>=6.0.12.2,<7.0.0.0',
 'wheel>=0.38.0,<0.39.0']

setup_kwargs = {
    'name': 'wtforglib',
    'version': '0.7.1',
    'description': 'A set of utility functions I use in my various python projects.',
    'long_description': '# wtforglib\n\n[![Build Status](https://github.com/wtfo-guru/wtforglib/workflows/Wtforglib/badge.svg?branch=main&event=push)](https://github.com/wtfo-guru/wtforglib/actions?query=workflow%3AWtforglib)\n[![codecov](https://codecov.io/gh/wtfo-guru/wtforglib/branch/main/graph/badge.svg)](https://codecov.io/gh/wtfo-guru/wtforglib)\n[![Python Version](https://img.shields.io/pypi/pyversions/wtforglib.svg)](https://pypi.org/project/wtforglib/)\n[![wemake-python-styleguide](https://img.shields.io/badge/style-wemake-000000.svg)](https://github.com/wemake-services/wemake-python-styleguide)\n\nA set of utility functions I use in my various python projects.\n\n## Features\n\n- Fully typed with annotations and checked with mypy, [PEP561 compatible](https://www.python.org/dev/peps/pep-0561/)\n- Add yours!\n\n\n## Installation\n\n```bash\npip install wtforglib\n```\n\n## Documentation\n\n- [Stable](https://wtforglib.readthedocs.io/en/stable)\n\n- [Latest](https://wtforglib.readthedocs.io/en/latest)\n\n## License\n\n[MIT](https://github.com/wtfo-guru/wtforglib/blob/main/LICENSE)\n\n## Credits\n\nThis project was generated with [`wemake-python-package`](https://github.com/wemake-services/wemake-python-package). Current template version is: [5fe13ee2646d5cf38736bacfa8f9dbbfac092efb](https://github.com/wemake-services/wemake-python-package/tree/5fe13ee2646d5cf38736bacfa8f9dbbfac092efb). See what is [updated](https://github.com/wemake-services/wemake-python-package/compare/5fe13ee2646d5cf38736bacfa8f9dbbfac092efb...main) since then.\n',
    'author': 'Quien Sabe',
    'author_email': 'qs5779@mail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/wtfo-guru/wtforglib',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
